import argparse
import networkx as nx
import itertools
from .rwc import get_rwc


def read_nodelist(filename: str) -> str:
	"""
	Defines an iterator over the nodes contained in a nodelist
	
	:param filename: filename to the nodelist
	:return: a node
	"""
	with open(filename, "r") as f:
		for row in f:
			yield row[0:-1]  # remove newline char
			

def main():
	parser = argparse.ArgumentParser()
	parser.add_argument(
		"edgelist",
		help="The path to the file containing the graph weighted edgelist. An example of row accepted format is 'node1,node2,weight\\n'",
		type=str
	)
	parser.add_argument(
		"community1_nodelist",
		help="The path to the nodelist containing the nodes belonging to the first side of the controversy",
		type=str
	)
	parser.add_argument(
		"community2_nodelist",
		help="The path to the nodelist containing the nodes belonging to the second side of the controversy",
		type=str
	)
	parser.add_argument(
		"percent",
		help="The percentage of nodes of each side to take as starting points for each simulation. This parameter is a float number, so a number between 0.0 and 1.0",
		type=float
	)
	parser.add_argument(
		"n",
		help="Total number of simulations to run",
		type=int
	)
	parser.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true")
	parser.add_argument("-l", "--log", help="print a log everytime a simulation has completed", action="store_true")
	args = parser.parse_args()
	
	edgelist_filename = args.edgelist
	community1_nodelist_filename = args.community1_nodelist
	community2_nodelist_filename = args.community2_nodelist
	percent = args.percent
	n = args.n
	summary = True if args.verbose else False
	completion_logs = True if args.log else False

	g = nx.read_weighted_edgelist(edgelist_filename, delimiter=',')

	side1 = [node for node in read_nodelist(community1_nodelist_filename)]
	side2 = [node for node in read_nodelist(community2_nodelist_filename)]

	rwc = get_rwc(g, side1, side2, percent, n, summary, completion_logs)
	if summary:
		# generic infos
		print(f"Nodes: {len(g.nodes())}")
		print(f"Community 1 nodes: {len(side1)}")
		print(f"Community 2 nodes: {len(side2)}")
		# rwc
		print(f"RWC score: {rwc['rwc_score']}")
		# frequency prints
		print("Frequency - Number of times the random walker started in a community and ended up in the same / another")
		for (side1, side2) in itertools.product(["side1", "side2"], ["side1", "side2"]):
			frequency = rwc["frequencies"][side1][side2]
			print(f"\t{side1} to {side2}: {frequency}")
		# probability prints
		print("Probability - Conditional probabilities that the random walker starts in community A and ends up in community B")
		for (side1, side2) in itertools.product(["community1", "community2"], ["community1", "community2"]):
			probability = rwc["probabilities"][side1][side2]
			print(f"\t{side1} to {side2}: {probability}")
	else:
		print(f"RWC: {rwc}")


if __name__ == "__main__":
	main()



import networkx as nx
import random
import concurrent.futures
from datetime import datetime

__all__ = ["get_rwc"]


def perform_random_walk(g: nx.DiGraph, starting_node: object, user_nodes_side1: list, user_nodes_side2: list) -> str:
	"""
	Starting from the specified starting node, performs a random walk and returns the first side that is reached.
	As noted by the authors in their original work, the random walker could get stuck (i.e., dangling vertices)

	:param g: a networkx directed Graph
	:param starting_node: object, the node in the graph to use as a starting point for the random walk
	:param user_nodes_side1: list, the nodes belonging to the first side
	:param user_nodes_side2: list, the nodes belonging to the second side
	:return: str returns "side1" if we ended up in a "side1" node, "side2" otherwise
	"""
	current_node = starting_node
	while True:
		current_node = random.choice([_ for _ in g.neighbors(current_node)])
		if current_node in user_nodes_side1:
			return "side1"
		if current_node in user_nodes_side2:
			return "side2"


def perform_simulation(g: nx.DiGraph, side1_nodes: list, side1_percent: int, side2_nodes: list, side2_percent: int) -> dict:
	"""
	Performs the specified number of random walks, for each side, and returns the frequencies of starting and ending
	in the same side and in different sides

	:param g: a networkx directed Graph
	:param side1_nodes: list, the nodes belonging to the first side
	:param side1_percent: float, between 0.0 and 1.0 is the percentage of nodes of each partition that we want to use
							a starting point in each simulation
	:param side2_nodes: list, the nodes belonging to the second side
	:param side2_percent: float, between 0.0 and 1.0 is the percentage of nodes of each partition that we want to use
							a starting point in each simulation
	:return: dictionary specifying for each side from where the random walk started (key) a dict (value) specifying a
			side (key) and the integer number of times (value) the random walk ended in that side
	"""
	results = {
		"side1": {
			"side1": 0,
			"side2": 0
		},
		"side2": {
			"side2": 0,
			"side1": 0
		}
	}

	user_nodes_side1 = [random.choice(side1_nodes) for _ in range(side1_percent)]  # starting point nodes from side 1
	user_nodes_side2 = [random.choice(side2_nodes) for _ in range(side2_percent)]  # starting point nodes from side 2

	# from side1 to side2
	for (index, start_node) in enumerate(user_nodes_side1):
		other_nodes = user_nodes_side1[:index] + user_nodes_side1[index + 1:]  # other nodes in that partition except the one just picked
		side = perform_random_walk(g, start_node, other_nodes, user_nodes_side2)  # get the side we endend after the random walk
		if side != "":
			results["side1"][side] += 1

	# from side2 to side1
	for (index, start_node) in enumerate(user_nodes_side2):
		other_nodes = user_nodes_side2[:index] + user_nodes_side2[index + 1:]
		side = perform_random_walk(g, start_node, user_nodes_side1, other_nodes)
		if side != "":
			results["side2"][side] += 1

	return results


def get_rwc(
		g: nx.DiGraph,
		side1: list,
		side2: list,
		percent: float,
		n: int,
		summary: bool = False,
		completion_logs: bool = False
) -> object:
	"""
	Given a graph and two partitions belonging to it, returns the Random Walk Controversy score of the network.
	The most straightforward way to compute RWC is via Monte Carlo sampling: since the random walks are independent of
	each other, they are executed in parallel.

	This code is the rewriting in python3 of the code made by the original author 
	(https://github.com/gvrkiran/controversy-detection)

	:param g: a networkx directed Graph
	:param side1: list, the nodes belonging to the first side
	:param side2: list, the nodes belonging to the second side
	:param percent: float, between 0.0 and 1.0 is the number of nodes of each partition that we want to use a staring point in each simulation
	:param n: int, total number of simulations to run
	:param summary: bool, optional if True the function returns a brief summary (a dictionary) about the computation 
					(e.g. the frequencies, probabilities, the rwc score), otherwise it just returns the rwc score
	:param completion_logs: bool, optional If True prints a log on stdout every time a simulation has completed
	:return: rwc score if summary is False, otherwise a dictionary containing statistics related to the simulations
	"""
	frequencies = {
		"side1": {
			"side1": 0,
			"side2": 0
		},
		"side2": {
			"side2": 0,
			"side1": 0
		}
	}

	side1_nodes = int(percent * len(side1))
	side2_nodes = int(percent * len(side2))
	
	# parallel execution of simulations
	simulations_completed = 0
	with concurrent.futures.ProcessPoolExecutor() as executor:
		futures = [executor.submit(perform_simulation, g, side1, side1_nodes, side2, side2_nodes) for i in range(0, n)]
		for future in concurrent.futures.as_completed(futures):
			simulation_frequencies = future.result()
			frequencies["side1"]["side1"] += simulation_frequencies["side1"]["side1"]
			frequencies["side1"]["side2"] += simulation_frequencies["side1"]["side2"]
			frequencies["side2"]["side1"] += simulation_frequencies["side2"]["side1"]
			frequencies["side2"]["side2"] += simulation_frequencies["side2"]["side2"]
			if completion_logs:
				simulations_completed += 1
				formatted_current_time = datetime.now().strftime("[%H:%M:%S]")
				print(f"{formatted_current_time} Simulation #{simulations_completed} completed")

	p_xx = frequencies["side1"]["side1"] * 1.0 / (frequencies["side1"]["side1"] + frequencies["side2"]["side1"])
	p_xy = frequencies["side1"]["side2"] * 1.0 / (frequencies["side1"]["side2"] + frequencies["side2"]["side2"])
	p_yx = frequencies["side2"]["side1"] * 1.0 / (frequencies["side1"]["side1"] + frequencies["side2"]["side1"])
	p_yy = frequencies["side2"]["side2"] * 1.0 / (frequencies["side1"]["side2"] + frequencies["side2"]["side2"])
	
	rwc_score = p_xx * p_yy - p_xy * p_yx
	if not summary:
		return rwc_score
	else:
		return {
			"rwc_score": rwc_score,
			"frequencies": frequencies,
			"probabilities": {
				"community1": {
					"community1": p_xx,
					"community2": p_xy
				},
				"community2": {
					"community1": p_yx,
					"community2": p_yy
				}
			}
		}





# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['random_walk_controversy']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.8.5,<3.0.0']

setup_kwargs = {
    'name': 'random-walk-controversy',
    'version': '0.1.2',
    'description': "Parallel implementation of Garimella's algorithm to get the RWC score of a network",
    'long_description': '# Random Walk Controversy\n\nThis repo contains a parallel implementation of Kiran Garimella\'s algorithm to compute the random walk controversy \nscore of a graph.  \n\nLet _G(V,E)_ be a graph with two partitions _X_ and _Y_; in the paper\n["Quantifying Controversy on Social Media"](https://dl.acm.org/doi/abs/10.1145/3140565) Garimella et al. define the \nRandom Walk Controversy (RWC) measure as follows: _"Consider two random walks, one ending in partition X and one ending \nin partition Y , RWC is the difference of the probabilities of two events: (i) both random walks started from the \npartition they ended in and (ii) both random walks started in a partition other than the one they ended in.”._  \nThe measure is quantified as $RWC = P_{XX}P_{YY} - P_{YX}P_{XY}$ where $P_{AB}$ is the conditional \nprobability $P_{AB} = Pr[\\mbox{start in partition }A \\mid \\mbox{end in partition }B]$.  \n\nSince the probabilities are computed by making simulations consisting in random walks and the simulations are independent\nof each other, they can easily be done in parallel.  \nThe following table shows the performance and results comparison made between the (sequential) \n[implementation](https://github.com/gvrkiran/controversy-detection) provided by one of the original authors and this \nimplementation. The datasets can be found in the same repo as the original authors\' implementation.\n\n| **Dataset**    | **Seq. time (s)** | **Seq. RWC score** | **Par. time (s)** | **Par. RWC score** | **Speedup** |\n|----------------|-------------------|--------------------|-------------------|--------------------|-------------|\n| baltimore      | 618               | 0.869              | 71                | 0.872              | 8.70        |\n| beefban        | 128               | 0.873              | 19                | 0.882              | 6.73        |\n| gunsense       | 2232              | 0.851              | 221               | 0.853              | 10.10       |\n| indiana        | 229               | 0.720              | 37                | 0.727              | 6.19        |\n| indiasdaughter | 490               | 0.825              | 62                | 0.832              | 7.90        |\n\nBeside the evident speedup obtained by exploiting a multicore architecture, we can see that the sequential and parallel\nversions almost converge to the same results.\n\n## Installation\nTo install the latest version of the library just download (or clone) the current project, open a terminal and run the \nfollowing commands:\n```\npip install -r requirements.txt\npip install .\n```\nAlternatively use pip\n```\npip install random-walk-controversy\n```\n\n## Usage\n### Command line interface\n```\npython3 -m random_walk_controversy [-h] [-v] [-l] edgelist community1_nodelist community2_nodelist percent n\n```\nMore info about the parameters can be fetched by using the ```-h``` option.  \nThe option  ```-v``` can be used to increase output verbosity and print, alongside the rwc score, the statistics about\nrandom walks. If not specified, only the RWC score is printed out.  \nFinally, the ```-l``` option displays a log on the terminal everytime a simulation is completed; I have found this option \npretty usefully since it allows to estimate the time for the algorithm to complete and understand if the algorithm got \nstuck.\n\n\n### Python library\nAfter the installation, it is possible to compute the rwc score directly in the python interpreter by using the function\n```get_rwc``` inside the ```random_walk_controversy``` package.\n\n#### Example\n```python\n>>> from random_walk_controversy import get_rwc\n>>> graph = read_edgelist()\n>>> side1_nodes = read_nodelist()  # list of nodes belonging to partition 1\n>>> side2_nodes = read_nodelist()  # list of nodes belonging to partition 2\n>>> node_percentage = 0.3\n>>> number_simulations = 1000\n>>> get_rwc(graph, side1_nodes, side2_nodes, node_percentage, number_simulations)\n76.233\n```\n',
    'author': 'Gianluca Morcaldi',
    'author_email': 'bendico765@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bendico765/random_walk_controversy',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
